from path import Path
import os, sys, asyncio

if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

from multiprocessing import Process

import pytest
import pytestqt
import cadquery as cq

from PyQt5.QtCore import Qt, QSettings, QPoint, QEvent, QSize
from PyQt5.QtWidgets import QFileDialog, QMessageBox
from PyQt5.QtGui import QMouseEvent

from cq_editor.__main__ import MainWindow
from cq_editor.widgets.editor import Editor
from cq_editor.cq_utils import export, get_occ_color

code = """import cadquery as cq
result = cq.Workplane("XY" )
result = result.box(3, 3, 0.5)
result = result.edges("|Z").fillet(0.125)"""

code_bigger_object = """import cadquery as cq
result = cq.Workplane("XY" )
result = result.box(20, 20, 0.5)
result = result.edges("|Z").fillet(0.125)
"""

code_show_Workplane = """import cadquery as cq
result = cq.Workplane("XY" )
result = result.box(3, 3, 0.5)
result = result.edges("|Z").fillet(0.125)

show_object(result)
"""

code_show_Workplane_named = """import cadquery as cq
result = cq.Workplane("XY" )
result = result.box(3, 3, 0.5)
result = result.edges("|Z").fillet(0.125)

log('test')
show_object(result,name='test')
"""

code_show_Shape = """import cadquery as cq
result = cq.Workplane("XY" )
result = result.box(3, 3, 0.5)
result = result.edges("|Z").fillet(0.125)

show_object(result.val())
"""

code_debug_Workplane = """import cadquery as cq
result = cq.Workplane("XY" )
result = result.box(3, 3, 0.5)
result = result.edges("|Z").fillet(0.125)

debug(result)
"""

code_multi = """import cadquery as cq
result1 = cq.Workplane("XY" ).box(3, 3, 0.5)
result2 = cq.Workplane("XY" ).box(3, 3, 0.5).translate((0,15,0))
"""

code_nested_top = """import test_nested_bottom
"""

code_nested_bottom = """a=1
"""

code_reload_issue = """wire0 = cq.Workplane().lineTo(5, 5).lineTo(10, 0).close().val()
solid1 = cq.Solid.extrudeLinear(cq.Face.makeFromWires(wire0), cq.Vector(0, 0, 1))
r1 = cq.Workplane(solid1).translate((10, 0, 0))
"""

code_show_all = """import cadquery as cq
b = cq.Workplane().box(1,1,1)
sh = b.val()
a = cq.Assembly().add(sh)
sk = cq.Sketch().rect(1,1)
"""


def _modify_file(code, path="test.py"):
    with open(path, "w", 1) as f:
        f.write(code)


def modify_file(code, path="test.py"):
    p = Process(target=_modify_file, args=(code, path))
    p.start()
    p.join()


def get_center(widget):

    pos = widget.pos()
    pos.setX(pos.x() + widget.width() // 2)
    pos.setY(pos.y() + widget.height() // 2)

    return pos


def get_bottom_left(widget):

    pos = widget.pos()
    pos.setY(pos.y() + widget.height())

    return pos


def get_rgba(ais):

    alpha = ais.Transparency()
    color = get_occ_color(ais)

    return color.redF(), color.greenF(), color.blueF(), alpha


@pytest.fixture
def main(qtbot, mocker):

    mocker.patch.object(QMessageBox, "question", return_value=QMessageBox.Yes)

    win = MainWindow()
    win.show()

    qtbot.addWidget(win)

    editor = win.components["editor"]
    editor.set_text(code)

    debugger = win.components["debugger"]
    debugger._actions["Run"][0].triggered.emit()

    return qtbot, win


@pytest.fixture
def main_clean(qtbot, mocker):

    mocker.patch.object(QMessageBox, "question", return_value=QMessageBox.Yes)

    win = MainWindow()
    win.show()

    qtbot.addWidget(win)

    with qtbot.waitExposed(win):
        editor = win.components["editor"]
        editor.set_text(code)

    return qtbot, win


@pytest.fixture
def main_clean_do_not_close(qtbot, mocker):

    mocker.patch.object(QMessageBox, "question", return_value=QMessageBox.No)

    win = MainWindow()
    win.show()

    qtbot.addWidget(win)

    with qtbot.waitExposed(win):
        editor = win.components["editor"]
        editor.set_text(code)

    return qtbot, win


@pytest.fixture
def main_multi(qtbot, mocker):

    mocker.patch.object(QMessageBox, "question", return_value=QMessageBox.Yes)
    mocker.patch.object(QFileDialog, "getSaveFileName", return_value=("out.step", ""))

    win = MainWindow()
    win.show()

    qtbot.addWidget(win)

    with qtbot.waitExposed(win):
        editor = win.components["editor"]
        editor.set_text(code_multi)

        debugger = win.components["debugger"]
        debugger._actions["Run"][0].triggered.emit()

    return qtbot, win


def test_render(main):

    qtbot, win = main

    obj_tree_comp = win.components["object_tree"]
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    console = win.components["console"]
    log = win.components["log"]

    # enable CQ reloading
    debugger.preferences["Reload CQ"] = True

    # check that object was rendered
    assert obj_tree_comp.CQ.childCount() == 1

    # check that object was removed
    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    # check that object was rendered usin explicit show_object call
    editor.set_text(code_show_Workplane)
    debugger._actions["Run"][0].triggered.emit()

    assert obj_tree_comp.CQ.childCount() == 1

    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    # check that cq.Shape object was rendered using explicit show_object call
    editor.set_text(code_show_Shape)
    debugger._actions["Run"][0].triggered.emit()

    assert obj_tree_comp.CQ.childCount() == 1

    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    # test rendering via console
    console.execute(code_show_Workplane)
    assert obj_tree_comp.CQ.childCount() == 1

    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    console.execute(code_show_Shape)
    assert obj_tree_comp.CQ.childCount() == 1

    # check object rendering using show_object call with a name specified and
    # debug call
    editor.set_text(code_show_Workplane_named)
    debugger._actions["Run"][0].triggered.emit()

    qtbot.wait(100)
    assert obj_tree_comp.CQ.child(0).text(0) == "test"
    assert "test" in log.toPlainText().splitlines()[-1]

    # cq reloading check
    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    editor.set_text(code_reload_issue)
    debugger._actions["Run"][0].triggered.emit()

    qtbot.wait(100)
    assert obj_tree_comp.CQ.childCount() == 3

    debugger._actions["Run"][0].triggered.emit()
    qtbot.wait(100)
    assert obj_tree_comp.CQ.childCount() == 3


def test_export(main, mocker):

    qtbot, win = main

    debugger = win.components["debugger"]
    debugger._actions["Run"][0].triggered.emit()

    # set focus
    obj_tree = win.components["object_tree"].tree
    obj_tree_comp = win.components["object_tree"]
    qtbot.mouseClick(obj_tree, Qt.LeftButton)
    qtbot.keyClick(obj_tree, Qt.Key_Down)
    qtbot.keyClick(obj_tree, Qt.Key_Down)

    # export STL
    mocker.patch.object(QFileDialog, "getSaveFileName", return_value=("out.stl", ""))
    obj_tree_comp._export_STL_action.triggered.emit()
    assert os.path.isfile("out.stl")

    # export STEP
    mocker.patch.object(QFileDialog, "getSaveFileName", return_value=("out.step", ""))
    obj_tree_comp._export_STEP_action.triggered.emit()
    assert os.path.isfile("out.step")

    # clean
    os.remove("out.step")
    os.remove("out.stl")


def number_visible_items(viewer):

    from OCP.AIS import AIS_ListOfInteractive

    l = AIS_ListOfInteractive()

    viewer_ctx = viewer._get_context()
    viewer_ctx.DisplayedObjects(l)

    return l.Extent()


def test_inspect(main):

    qtbot, win = main

    # set focus and make invisible
    obj_tree = win.components["object_tree"].tree
    qtbot.mouseClick(obj_tree, Qt.LeftButton)
    qtbot.keyClick(obj_tree, Qt.Key_Down)
    qtbot.keyClick(obj_tree, Qt.Key_Down)
    qtbot.keyClick(obj_tree, Qt.Key_Space)

    # enable object inspector
    insp = win.components["cq_object_inspector"]
    insp._toolbar_actions[0].toggled.emit(True)

    # check if all stack items are visible in the tree
    assert insp.root.childCount() == 3

    # check if correct number of items is displayed
    viewer = win.components["viewer"]

    insp.setCurrentItem(insp.root.child(0))
    assert number_visible_items(viewer) == 4

    insp.setCurrentItem(insp.root.child(1))
    assert number_visible_items(viewer) == 7

    insp.setCurrentItem(insp.root.child(2))
    assert number_visible_items(viewer) == 4

    insp._toolbar_actions[0].toggled.emit(False)
    assert number_visible_items(viewer) == 3


class event_loop(object):
    """Used to mock the QEventLoop for the debugger component"""

    def __init__(self, callbacks):

        self.callbacks = callbacks
        self.i = 0

    def exec_(self):

        if self.i < len(self.callbacks):
            self.callbacks[self.i]()
            self.i += 1

    def exit(self, *args):

        pass


def patch_debugger(debugger, event_loop_mock):

    debugger.inner_event_loop.exec_ = event_loop_mock.exec_
    debugger.inner_event_loop.exit = event_loop_mock.exit


def test_debug(main, mocker):

    # store the tracing function
    trace_function = sys.gettrace()

    def assert_func(x):
        """Neddedd to perform asserts in lambdas"""
        assert x

    qtbot, win = main

    # clear all
    obj_tree = win.components["object_tree"]
    obj_tree.toolbarActions()[0].triggered.emit()

    editor = win.components["editor"]
    editor.set_text(code)

    debugger = win.components["debugger"]
    actions = debugger._actions["Run"]
    run, debug, step, step_in, cont = actions

    variables = win.components["variables_viewer"]

    traceback_view = win.components["traceback_viewer"]

    def check_no_error_occured():
        """check that no error occured while stepping through the debugger"""
        assert "" == traceback_view.current_exception.text()

    viewer = win.components["viewer"]
    assert number_visible_items(viewer) == 3

    # check breakpoints
    assert debugger.set_breakpoints([])

    # check _frames
    assert debugger._frames == []

    # test step through
    ev = event_loop(
        [
            lambda: (
                assert_func(variables.model().rowCount() == 5),
                assert_func(number_visible_items(viewer) == 3),
                step.triggered.emit(),
            ),
            lambda: (
                assert_func(variables.model().rowCount() == 5),
                assert_func(number_visible_items(viewer) == 3),
                step.triggered.emit(),
            ),
            lambda: (
                assert_func(variables.model().rowCount() == 6),
                assert_func(number_visible_items(viewer) == 3),
                step.triggered.emit(),
            ),
            lambda: (
                assert_func(variables.model().rowCount() == 6),
                assert_func(number_visible_items(viewer) == 4),
                cont.triggered.emit(),
            ),
        ]
    )

    patch_debugger(debugger, ev)

    debug.triggered.emit(True)

    check_no_error_occured()

    assert variables.model().rowCount() == 2
    assert number_visible_items(viewer) == 4

    # test exit debug
    ev = event_loop(
        [
            lambda: (step.triggered.emit(),),
            lambda: (
                assert_func(variables.model().rowCount() == 5),
                assert_func(number_visible_items(viewer) == 3),
                debug.triggered.emit(False),
            ),
        ]
    )

    patch_debugger(debugger, ev)

    debug.triggered.emit(True)

    check_no_error_occured()

    assert variables.model().rowCount() == 1
    assert number_visible_items(viewer) == 3

    # test breakpoint
    ev = event_loop(
        [
            lambda: (cont.triggered.emit(),),
            lambda: (
                assert_func(variables.model().rowCount() == 6),
                assert_func(number_visible_items(viewer) == 4),
                cont.triggered.emit(),
            ),
        ]
    )

    patch_debugger(debugger, ev)

    editor.debugger.set_breakpoints([(4, None)])

    debug.triggered.emit(True)

    check_no_error_occured()

    assert variables.model().rowCount() == 2
    assert number_visible_items(viewer) == 4

    # test breakpoint without using singals
    ev = event_loop(
        [
            lambda: (cont.triggered.emit(),),
            lambda: (
                assert_func(variables.model().rowCount() == 6),
                assert_func(number_visible_items(viewer) == 4),
                cont.triggered.emit(),
            ),
        ]
    )

    patch_debugger(debugger, ev)

    editor.debugger.set_breakpoints([(4, None)])

    debugger.debug(True)

    check_no_error_occured()

    assert variables.model().rowCount() == 2
    assert number_visible_items(viewer) == 4

    # test debug() without using singals
    ev = event_loop(
        [
            lambda: (cont.triggered.emit(),),
            lambda: (
                assert_func(variables.model().rowCount() == 6),
                assert_func(number_visible_items(viewer) == 4),
                cont.triggered.emit(),
            ),
        ]
    )

    patch_debugger(debugger, ev)

    editor.set_text(code_debug_Workplane)
    editor.debugger.set_breakpoints([(4, None)])

    debugger.debug(True)

    check_no_error_occured()

    CQ = obj_tree.CQ

    # object 1 (defualt color)
    r, g, b, a = get_rgba(CQ.child(0).ais)
    # assert a == pytest.approx(0.2)
    assert r == 1.0

    assert variables.model().rowCount() == 2
    assert number_visible_items(viewer) == 4

    # restore the tracing function
    sys.settrace(trace_function)


code_err1 = """import cadquery as cq
(
result = cq.Workplane("XY" ).box(3, 3, 0.5).edges("|Z").fillet(0.125)
"""

code_err2 = """import cadquery as cq
result = cq.Workplane("XY" ).box(3, 3, 0.5).edges("|Z").fillet(0.125)
f()
"""
code_err3 = """import cadquery as cq
result = cq.Workplane("XY" ).box(3, 3, 0)
"""

base_editor_text = """import cadquery as cq
result = cq.Workplane().box(10, 10, 10)
"""


def test_traceback(main):

    # store the tracing function
    trace_function = sys.gettrace()

    qtbot, win = main

    editor = win.components["editor"]
    debugger = win.components["debugger"]
    traceback_view = win.components["traceback_viewer"]

    actions = debugger._actions["Run"]
    run, debug, step, step_in, cont = actions

    editor.set_text(code_err1)
    run.triggered.emit()

    assert "SyntaxError" in traceback_view.current_exception.text()

    debug.triggered.emit()

    assert "SyntaxError" in traceback_view.current_exception.text()
    assert debug.isChecked() == False

    editor.set_text(code_err2)
    run.triggered.emit()

    assert "NameError" in traceback_view.current_exception.text()
    assert hasattr(sys, "last_traceback")

    del sys.last_traceback
    assert not hasattr(sys, "last_traceback")

    # test last_traceback with debug
    ev = event_loop([lambda: (cont.triggered.emit(),)])
    patch_debugger(debugger, ev)

    debugger.debug(True)

    assert "NameError" in traceback_view.current_exception.text()
    assert hasattr(sys, "last_traceback")
    assert traceback_view.tree.root.childCount() == 1

    # restore the tracing function
    sys.settrace(trace_function)

    # check if errors deeper in CQ are reported too
    editor.set_text(code_err3)
    run.triggered.emit()

    assert "Standard_DomainError" in traceback_view.current_exception.text()
    assert traceback_view.tree.root.childCount() == 3  # 1 in user code + 2 in CQ code


@pytest.fixture
def editor(qtbot):

    win = Editor()
    win.show()

    qtbot.addWidget(win)

    return qtbot, win


def conv_line_ends(text):

    return "\n".join(text.splitlines())


def test_editor(monkeypatch, editor):

    qtbot, editor = editor

    with open("test.py", "w") as f:
        f.write(code)

    # check that no text is present
    assert editor.get_text_with_eol() == ""

    # check that loading from file works properly
    editor.load_from_file("test.py")
    assert len(editor.get_text_with_eol()) > 0
    assert conv_line_ends(editor.get_text_with_eol()) == code

    # check that loading from file works properly
    editor.new()
    assert editor.get_text_with_eol() == ""

    # monkeypatch QFileDialog methods
    def filename(*args, **kwargs):
        return "test.py", None

    def filename2(*args, **kwargs):
        return "test2.py", None

    monkeypatch.setattr(QFileDialog, "getOpenFileName", staticmethod(filename))

    monkeypatch.setattr(QFileDialog, "getSaveFileName", staticmethod(filename2))

    # check that open file works properly
    editor.open()
    assert conv_line_ends(editor.get_text_with_eol()) == code

    # check that save file works properly
    editor.new()
    qtbot.mouseClick(editor, Qt.LeftButton)
    qtbot.keyClick(editor, Qt.Key_A)

    assert editor.document().isModified() == True

    editor.filename = "test2.py"
    editor.save()

    assert editor.document().isModified() == False

    monkeypatch.setattr(QFileDialog, "getOpenFileName", staticmethod(filename2))

    editor.open()
    assert editor.get_text_with_eol() == "a"

    # check that save as works properly
    os.remove("test2.py")
    editor.save_as()
    assert os.path.exists(filename2()[0])

    # test persistance
    settings = QSettings("test")
    editor.saveComponentState(settings)

    editor.new()
    assert editor.get_text_with_eol() == ""

    editor.restoreComponentState(settings)
    assert editor.get_text_with_eol() == "a"

    # test error handling
    os.remove("test2.py")
    assert not os.path.exists("test2.py")
    editor.restoreComponentState(settings)


def test_size_hint(editor):
    """
    Tests the ability to get the size hit from the code editor widget.
    """
    qtbot, editor = editor

    size_hint = editor.sizeHint()

    assert size_hint == QSize(256, 192)


def test_clear_selection(editor):
    """
    Tests the ability to clear selected text.
    """
    qtbot, editor = editor

    # Set a block of text and make sure it is visible
    editor.set_text(base_editor_text)
    editor.document().setModified(False)
    assert editor.get_text_with_eol() == base_editor_text

    # Remove all the text and make sure it was removed
    editor.selectAll()
    cursor = editor.textCursor()
    cursor.removeSelectedText()
    editor.setTextCursor(cursor)
    editor.document().setModified(False)
    assert editor.get_text_with_eol() == ""

    # Test the ability to deselect a selected area
    editor.set_text(base_editor_text)
    editor.selectAll()
    assert editor.get_selection_range() == (0, 2)
    editor.clear_selection()
    assert editor.get_selection_range() == (0, 0)


def test_insert_remove_line_start(editor):
    """
    Tests the ability to remove and insert characters from/to the beginning of a line.
    """
    qtbot, editor = editor

    # Set a block of text and make sure it is visible
    editor.set_text(base_editor_text)
    editor.insert_line_start("# ", 0)
    editor.document().setModified(False)
    assert editor.get_text_with_eol() == "# " + base_editor_text

    # Remove the comment character from the line
    editor.remove_line_start("# ", 0)
    editor.document().setModified(False)
    assert editor.get_text_with_eol() == base_editor_text


def test_indent_unindent(editor):
    """
    Check to make sure that indent and un-indent work properly.
    """
    qtbot, editor = editor

    # Set the base text
    editor.set_text(base_editor_text)

    # Indent the text and check
    editor.selectAll()
    qtbot.keyClick(editor, Qt.Key_Tab)
    editor.document().setModified(False)
    assert editor.get_text_with_eol() != base_editor_text

    # Unindent the code with a direct method call and check
    editor.selectAll()
    start_line, end_line = editor.get_selection_range()
    # +1 here to compesate for how black wants the multi-line string
    editor.undo_indent(list(range(start_line, end_line + 1)))
    editor.document().setModified(False)
    assert editor.get_text_with_eol() == base_editor_text

    # Indent the code again with a direct method call and check
    editor.selectAll()
    start_line, end_line = editor.get_selection_range()
    editor.do_indent(list(range(start_line, end_line)))
    editor.document().setModified(False)
    assert editor.get_text_with_eol() != base_editor_text

    # Unindent the code again with a keystroke
    editor.selectAll()
    qtbot.keyClick(editor, Qt.Key_Backtab)
    editor.document().setModified(False)
    assert editor.get_text_with_eol() == base_editor_text

    # Indent just the second line
    editor.clear_selection()
    editor.do_indent([1])
    assert editor.get_text_with_eol() != base_editor_text


def test_set_color_scheme(editor):
    """
    Make sure that the color theme can be switched without error.
    """
    qtbot, editor = editor

    editor.set_color_scheme("Light")
    editor.set_color_scheme("Dark")


def test_go_to_line(editor):
    """
    Tests to make sure the caller can set the current line of the code.
    """
    qtbot, editor = editor

    # Set the base text
    editor.set_text(base_editor_text)

    # Make sure the line changes
    editor.go_to_line(1)
    cursor = editor.textCursor()
    block = cursor.block()
    assert (block.blockNumber() + 1) == 1
    editor.go_to_line(2)
    cursor = editor.textCursor()
    block = cursor.block()
    assert (block.blockNumber() + 1) == 2


def test_toggle_comment(editor):
    """
    Tests to make sure that lines can be commented/uncommented.
    """
    qtbot, editor = editor

    # Set the base text
    editor.set_text(base_editor_text)

    # Try commenting and uncommenting a single line
    editor.go_to_line(1)
    editor.toggle_comment()
    assert editor.get_text_with_eol() != base_editor_text
    editor.toggle_comment()
    assert editor.get_text_with_eol() == base_editor_text

    # Try commenting and uncommenting multiple lines
    editor.selectAll()
    editor.toggle_comment()
    assert editor.get_text_with_eol() != base_editor_text
    editor.selectAll()
    editor.toggle_comment()
    assert editor.get_text_with_eol() == base_editor_text


def test_highlight_current_line(editor):
    """
    Make sure the current line can be highlighted without error.
    """
    qtbot, editor = editor

    # Set the base text
    editor.set_text(base_editor_text)

    # Highlight the first line
    editor.go_to_line(1)
    editor.highlight_current_line()


def test_set_remove_breakpoints(editor):
    """
    Make sure the breakpoints can be added and removed without error.
    """
    qtbot, editor = editor

    # Set the base text
    editor.set_text(base_editor_text)

    # Toggle breakpoints and check that they are there
    assert not editor.line_has_breakpoint(2)
    editor.toggle_breakpoint(2)
    assert editor.line_has_breakpoint(2)
    editor.toggle_breakpoint(2)
    assert not editor.line_has_breakpoint(2)


def test_search(editor):
    """
    Tests the search functionality.
    """
    qtbot, editor = editor

    # Set the base text
    editor.set_text(base_editor_text)

    # Test with no search match
    editor.search_widget.on_search_text_changed("~")
    assert editor.search_widget.match_label.text() == "0 matches"

    # Check to see that various search texts change the search controls
    editor.search_widget.on_search_text_changed("cq")
    assert editor.search_widget.match_label.text() == "1 of 2"

    # Make sure advancing to the next and previous matches works properly
    editor.search_widget.find_next()
    assert editor.search_widget.match_label.text() == "2 of 2"
    editor.search_widget.find_previous()
    assert editor.search_widget.match_label.text() == "1 of 2"

    # Make sure the show and hide search works
    editor.search_widget.show_search()
    assert editor.search_widget.isVisible()
    editor.search_widget.hide_search()
    assert not editor.search_widget.isVisible()

    # Test hotkeys
    qtbot.keyClick(editor, Qt.Key_F, modifier=Qt.ControlModifier)
    assert editor.search_widget.isVisible()
    qtbot.keyClick(editor, Qt.Key_F3)
    qtbot.keyClick(editor, Qt.Key_F3, modifier=Qt.AltModifier)


def test_line_number_area(editor):
    """
    Tests to make sure the line number area on the left of the editor is working correctly.
    """
    qtbot, editor = editor

    # Set the base text
    editor.set_text(base_editor_text)

    # Make sure the size hint can be retrieved without error
    editor.line_number_area.sizeHint()

    # Try to simulate a mouse click in the line number area
    pos = QPoint(10, 10)
    qtbot.mouseClick(editor.line_number_area, Qt.LeftButton, pos=pos)


def test_console(main):
    qtbot, win = main

    console = win.components["console"]

    # test execute_command
    a = []
    console.push_vars({"a": a})
    console.execute_command("a.append(1)")
    assert len(a) == 1

    # test print_text
    pos_orig = console._prompt_pos
    console.print_text("a")
    assert console._prompt_pos == pos_orig + len("a")

    # test reset of the console contents
    console.reset_console()
    assert len(console._control.toPlainText()) <= 10


def test_viewer(main):

    qtbot, win = main

    viewer = win.components["viewer"]

    # not sure how to test this, so only smoke tests

    # trigger all 'View' actions
    actions = viewer._actions["View"]
    for a in actions:
        a.trigger()


code_module = """def dummy(): return True"""

code_import = """from module import dummy
assert(dummy())"""


def test_module_import(main):

    qtbot, win = main

    editor = win.components["editor"]
    debugger = win.components["debugger"]
    traceback_view = win.components["traceback_viewer"]

    # save the dummy module
    with open("module.py", "w") as f:
        f.write(code_module)

    # run the code importing this module
    editor.set_text(code_import)
    debugger._actions["Run"][0].triggered.emit()

    # verify that no exception was generated
    assert traceback_view.current_exception.text() == ""


def test_auto_fit_view(main_clean):

    def concat(eye, proj, scale):
        return eye + proj + (scale,)

    def approx_view_properties(eye, proj, scale):

        return pytest.approx(eye + proj + (scale,))

    qtbot, win = main_clean

    editor = win.components["editor"]
    debugger = win.components["debugger"]
    viewer = win.components["viewer"]
    object_tree = win.components["object_tree"]

    view = viewer.canvas.view
    viewer.preferences["Fit automatically"] = False
    eye0, proj0, scale0 = view.Eye(), view.Proj(), view.Scale()
    # check if camera position is adjusted automatically when rendering for the
    # first time
    debugger.render()
    eye1, proj1, scale1 = view.Eye(), view.Proj(), view.Scale()
    assert concat(eye0, proj0, scale0) != approx_view_properties(eye1, proj1, scale1)

    # check if camera position is not changed fter code change
    editor.set_text(code_bigger_object)
    debugger.render()
    eye2, proj2, scale2 = view.Eye(), view.Proj(), view.Scale()
    assert concat(eye1, proj1, scale1) == approx_view_properties(eye2, proj2, scale2)

    # check if position is adjusted automatically after erasing all objects
    object_tree.removeObjects()
    debugger.render()
    eye3, proj3, scale3 = view.Eye(), view.Proj(), view.Scale()
    assert concat(eye2, proj2, scale2) != approx_view_properties(eye3, proj3, scale3)

    # check if position is adjusted automatically if settings are changed
    viewer.preferences["Fit automatically"] = True
    editor.set_text(code)
    debugger.render()
    eye4, proj4, scale4 = view.Eye(), view.Proj(), view.Scale()
    assert concat(eye3, proj3, scale3) != approx_view_properties(eye4, proj4, scale4)


def test_preserve_properties(main):
    qtbot, win = main

    debugger = win.components["debugger"]
    debugger._actions["Run"][0].triggered.emit()

    object_tree = win.components["object_tree"]
    object_tree.preferences["Preserve properties on reload"] = True

    assert object_tree.CQ.childCount() == 1
    props = object_tree.CQ.child(0).properties
    props["Visible"] = False
    # props["Color"] = "#caffee"
    # props["Alpha"] = 0.5

    debugger._actions["Run"][0].triggered.emit()

    assert object_tree.CQ.childCount() == 1
    props = object_tree.CQ.child(0).properties
    assert props["Visible"] == False
    # assert props["Color"].name() == "#caffee"
    # assert props["Alpha"] == 0.5


def test_selection(main_multi, mocker):

    qtbot, win = main_multi

    viewer = win.components["viewer"]
    object_tree = win.components["object_tree"]

    CQ = object_tree.CQ
    obj1 = CQ.child(0)
    obj2 = CQ.child(1)

    # export with two selected objects
    obj1.setSelected(True)
    obj2.setSelected(True)

    object_tree._export_STEP_action.triggered.emit()
    imported = cq.importers.importStep("out.step")
    assert len(imported.solids().vals()) == 2

    # export with one selected objects
    obj2.setSelected(False)

    object_tree._export_STEP_action.triggered.emit()
    imported = cq.importers.importStep("out.step")
    assert len(imported.solids().vals()) == 1

    # export with one selected objects
    obj1.setSelected(False)
    CQ.setSelected(True)

    object_tree._export_STEP_action.triggered.emit()
    imported = cq.importers.importStep("out.step")
    assert len(imported.solids().vals()) == 2

    # check if viewer and object tree are properly connected
    CQ.setSelected(False)
    obj1.setSelected(True)
    obj2.setSelected(True)
    ctx = viewer._get_context()
    ctx.InitSelected()
    shapes = []

    while ctx.MoreSelected():
        shapes.append(ctx.SelectedShape())
        ctx.NextSelected()
    assert len(shapes) == 2

    viewer.fit()
    qtbot.mouseClick(viewer.canvas, Qt.LeftButton)

    assert len(object_tree.tree.selectedItems()) == 0

    viewer.sigObjectSelected.emit([obj1.shape_display.wrapped])
    assert len(object_tree.tree.selectedItems()) == 1

    # go through different handleSelection paths
    qtbot.mouseClick(object_tree.tree, Qt.LeftButton)
    qtbot.keyClick(object_tree.tree, Qt.Key_Down)
    qtbot.keyClick(object_tree.tree, Qt.Key_Down)
    qtbot.keyClick(object_tree.tree, Qt.Key_Down)
    qtbot.keyClick(object_tree.tree, Qt.Key_Down)

    assert object_tree._export_STL_action.isEnabled() == False
    assert object_tree._export_STEP_action.isEnabled() == False
    assert object_tree._clear_current_action.isEnabled() == False
    assert object_tree.properties_editor.isEnabled() == False


def test_closing(main_clean_do_not_close):

    qtbot, win = main_clean_do_not_close

    editor = win.components["editor"]

    # make sure that windows is visible
    assert win.isVisible()

    # should not quit
    win.close()
    assert win.isVisible()

    # should quit
    editor.reset_modified()
    win.close()
    assert not win.isVisible()


def test_check_for_updates(main, mocker):

    qtbot, win = main

    # patch requests
    import requests

    mocker.patch.object(
        requests.models.Response,
        "json",
        return_value=[{"tag_name": "0.0.2", "draft": False}],
    )

    # stub QMessageBox about
    about_stub = mocker.stub()
    mocker.patch.object(QMessageBox, "about", about_stub)

    import cadquery

    cadquery.__version__ = "0.0.1"
    win.check_for_cq_updates()
    assert about_stub.call_args[0][1] == "Updates available"

    cadquery.__version__ = "0.0.3"
    win.check_for_cq_updates()
    assert about_stub.call_args[0][1] == "No updates available"


@pytest.mark.skipif(
    sys.platform.startswith("linux"), reason="Segfault workaround for linux"
)
def test_screenshot(main, mocker):

    qtbot, win = main

    mocker.patch.object(QFileDialog, "getSaveFileName", return_value=("out.png", ""))

    viewer = win.components["viewer"]
    viewer._actions["Tools"][0].triggered.emit()

    assert os.path.exists("out.png")


def test_resize(main):

    qtbot, win = main
    editor = win.components["editor"]

    editor.hide()
    qtbot.wait(50)
    editor.show()
    qtbot.wait(50)


code_simple_step = """import cadquery as cq
imported = cq.importers.importStep('shape.step')
"""


def test_relative_references(main):

    # create code with a relative reference in a subdirectory
    p = Path("test_relative_references")
    p.mkdir_p()
    p_code = p.joinpath("code.py")
    p_code.write_text(code_simple_step)
    # create the referenced step file
    shape = cq.Workplane("XY").box(1, 1, 1)
    p_step = p.joinpath("shape.step")
    export(shape, "step", p_step)
    # open code
    qtbot, win = main
    editor = win.components["editor"]
    editor.load_from_file(p_code)
    # render
    debugger = win.components["debugger"]
    debugger._actions["Run"][0].triggered.emit()
    # assert no errors
    traceback_view = win.components["traceback_viewer"]
    assert traceback_view.current_exception.text() == ""
    # assert one object has been rendered
    obj_tree_comp = win.components["object_tree"]
    assert obj_tree_comp.CQ.childCount() == 1
    # clean up
    p_code.remove_p()
    p_step.remove_p()
    p.rmdir_p()


code_color = """
import cadquery as cq
result = cq.Workplane("XY" ).box(1, 1, 1)

show_object(result, name ='1')
show_object(result, name ='2', options=dict(alpha=0.5,color='red'))
show_object(result, name ='3', options=dict(alpha=0.5,color='#ff0000'))
show_object(result, name ='4', options=dict(alpha=0.5,color=(255,0,0)))
show_object(result, name ='5', options=dict(alpha=0.5,color=(1.,0,0)))
show_object(result, name ='6', options=dict(rgba=(1.,0,0,.5)))
show_object(result, name ='7', options=dict(color=('ff','cc','dd')))
"""


def test_render_colors(main_clean):

    qtbot, win = main_clean

    obj_tree = win.components["object_tree"]
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    log = win.components["log"]

    editor.set_text(code_color)
    debugger._actions["Run"][0].triggered.emit()

    CQ = obj_tree.CQ

    # object 1 (defualt color)
    assert not CQ.child(0).ais.HasColor()

    # object 2
    r, g, b, a = get_rgba(CQ.child(1).ais)
    assert a == 0.5
    assert r == 1.0
    assert g == 0.0

    # object 3
    r, g, b, a = get_rgba(CQ.child(2).ais)
    assert a == 0.5
    assert r == 1.0

    # object 4
    r, g, b, a = get_rgba(CQ.child(3).ais)
    assert a == 0.5
    assert r == 1.0

    # object 5
    r, g, b, a = get_rgba(CQ.child(4).ais)
    assert a == 0.5
    assert r == 1.0

    # object 6
    r, g, b, a = get_rgba(CQ.child(5).ais)
    assert a == 0.5
    assert r == 1.0

    # check if error occured
    qtbot.wait(100)
    assert "Unknown color format" in log.toPlainText().splitlines()[-1]


def test_render_colors_console(main_clean):

    qtbot, win = main_clean

    obj_tree = win.components["object_tree"]
    log = win.components["log"]
    console = win.components["console"]

    console.execute_command(code_color)

    CQ = obj_tree.CQ

    # object 1 (defualt color)
    assert not CQ.child(0).ais.HasColor()

    # object 2
    r, g, b, a = get_rgba(CQ.child(1).ais)
    assert a == 0.5
    assert r == 1.0

    # object 3
    r, g, b, a = get_rgba(CQ.child(2).ais)
    assert a == 0.5
    assert r == 1.0

    # object 4
    r, g, b, a = get_rgba(CQ.child(3).ais)
    assert a == 0.5
    assert r == 1.0

    # object 5
    r, g, b, a = get_rgba(CQ.child(4).ais)
    assert a == 0.5
    assert r == 1.0

    # object 6
    r, g, b, a = get_rgba(CQ.child(5).ais)
    assert a == 0.5
    assert r == 1.0

    # check if error occured
    qtbot.wait(100)
    assert "Unknown color format" in log.toPlainText().splitlines()[-1]


code_shading = """
import cadquery as cq

res1 = cq.Workplane('XY').box(5, 7, 5)
res2 = cq.Workplane('XY').box(8, 5, 4)
show_object(res1)
show_object(res2,options={"alpha":0})
"""


def test_shading_aspect(main_clean):

    qtbot, win = main_clean

    obj_tree = win.components["object_tree"]
    editor = win.components["editor"]
    debugger = win.components["debugger"]

    editor.set_text(code_shading)
    debugger._actions["Run"][0].triggered.emit()

    CQ = obj_tree.CQ

    # get material aspects
    ma1 = CQ.child(0).ais.Attributes().ShadingAspect().Material()
    ma2 = CQ.child(1).ais.Attributes().ShadingAspect().Material()

    # verify that they are the same
    assert ma1.Shininess() == ma2.Shininess()


def test_confirm_new(monkeypatch, editor):

    qtbot, editor = editor

    # check that initial state is as expected
    assert editor.modified == False

    editor.document().setPlainText(code)
    assert editor.modified == True

    # monkeypatch the confirmation dialog and run both scenarios
    def cancel(*args, **kwargs):
        return QMessageBox.No

    def ok(*args, **kwargs):
        return QMessageBox.Yes

    monkeypatch.setattr(QMessageBox, "question", staticmethod(cancel))

    editor.new()
    assert editor.modified == True
    assert conv_line_ends(editor.get_text_with_eol()) == code

    monkeypatch.setattr(QMessageBox, "question", staticmethod(ok))

    editor.new()
    assert editor.modified == False
    assert editor.get_text_with_eol() == ""


code_show_topods = """
import cadquery as cq
result = cq.Workplane("XY" ).box(1, 1, 1)

show_object(result.val().wrapped)
"""


def test_render_topods(main):

    qtbot, win = main

    obj_tree_comp = win.components["object_tree"]
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    console = win.components["console"]

    # check that object was rendered
    assert obj_tree_comp.CQ.childCount() == 1

    # check that object was removed
    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    # check that object was rendered usin explicit show_object call
    editor.set_text(code_show_topods)
    debugger._actions["Run"][0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 1

    # test rendering of topods object via console
    console.execute("show(result.val().wrapped)")
    assert obj_tree_comp.CQ.childCount() == 2

    # test rendering of list of topods object via console
    console.execute("show([result.val().wrapped,result.val().wrapped])")
    assert obj_tree_comp.CQ.childCount() == 3


code_show_shape_list = """
import cadquery as cq
result1 = cq.Workplane("XY" ).box(1, 1, 1).val()
result2 = cq.Workplane("XY",origin=(0,1,1)).box(1, 1, 1).val()

show_object(result1)
show_object([result1,result2])
"""


def test_render_shape_list(main):

    qtbot, win = main

    log = win.components["log"]

    obj_tree_comp = win.components["object_tree"]
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    console = win.components["console"]

    # check that object was removed
    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    # check that object was rendered usin explicit show_object call
    editor.set_text(code_show_shape_list)
    debugger._actions["Run"][0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 2

    # test rendering of Shape via console
    console.execute("show(result1)")
    console.execute("show([result1,result2])")
    assert obj_tree_comp.CQ.childCount() == 4

    # smoke test exception in show
    console.execute('show("a")')


code_show_assy = """import cadquery as cq
result1 = cq.Workplane("XY" ).box(3, 3, 0.5)
assy = cq.Assembly()
assy.add(result1)

show_object(assy)
"""


def test_render_assy(main):

    qtbot, win = main

    obj_tree_comp = win.components["object_tree"]
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    console = win.components["console"]

    # check that object was removed
    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    # check that object was rendered usin explicit show_object call
    editor.set_text(code_show_assy)
    debugger._actions["Run"][0].triggered.emit()
    qtbot.wait(500)
    assert obj_tree_comp.CQ.childCount() == 1

    # test rendering via console
    console.execute("show(assy)")
    qtbot.wait(500)
    assert obj_tree_comp.CQ.childCount() == 2


code_show_ais = """import cadquery as cq
from cadquery.occ_impl.assembly import toCAF

import OCP

result1 = cq.Workplane("XY" ).box(3, 3, 0.5)
assy = cq.Assembly()
assy.add(result1)

lab, doc = toCAF(assy)
ais = OCP.XCAFPrs.XCAFPrs_AISObject(lab)

show_object(ais)
"""


def test_render_ais(main):

    qtbot, win = main

    obj_tree_comp = win.components["object_tree"]
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    console = win.components["console"]

    # check that object was removed
    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    # check that object was rendered usin explicit show_object call
    editor.set_text(code_show_ais)
    debugger._actions["Run"][0].triggered.emit()
    qtbot.wait(500)
    assert obj_tree_comp.CQ.childCount() == 1

    # test rendering via console
    console.execute("show(ais)")
    qtbot.wait(500)
    assert obj_tree_comp.CQ.childCount() == 2


code_show_sketch = """import cadquery as cq

s1 = cq.Sketch().rect(1,1)
s2 = cq.Sketch().segment((0,0), (0,3.),"s1")

show_object(s1)
show_object(s2)
"""


def test_render_sketch(main):

    qtbot, win = main

    obj_tree_comp = win.components["object_tree"]
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    console = win.components["console"]

    # check that object was removed
    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    # check that object was rendered usin explicit show_object call
    editor.set_text(code_show_sketch)
    debugger._actions["Run"][0].triggered.emit()
    qtbot.wait(500)
    assert obj_tree_comp.CQ.childCount() == 2

    # test rendering via console
    console.execute("show(s1); show(s2)")
    qtbot.wait(500)
    assert obj_tree_comp.CQ.childCount() == 4


def test_window_title(monkeypatch, main):

    fname = "test_window_title.py"

    with open(fname, "w") as f:
        f.write(code)

    qtbot, win = main

    # monkeypatch QFileDialog methods
    def filename(*args, **kwargs):
        return fname, None

    monkeypatch.setattr(QFileDialog, "getOpenFileName", staticmethod(filename))

    win.components["editor"].open()
    assert win.windowTitle().endswith(fname)

    # handle a new file
    win.components["editor"].new()
    # I don't really care what the title is, as long as it's not a filename
    assert not win.windowTitle().endswith(".py")


def test_module_discovery(tmp_path, editor):

    qtbot, editor = editor
    with open(tmp_path.joinpath("main.py"), "w") as f:
        f.write("import b")

    assert editor.get_imported_module_paths(str(tmp_path.joinpath("main.py"))) == []

    tmp_path.joinpath("b.py").touch()

    assert editor.get_imported_module_paths(str(tmp_path.joinpath("main.py"))) == [
        str(tmp_path.joinpath("b.py"))
    ]


def test_launch_syntax_error(tmp_path):

    # verify app launches when input file is bad
    win = MainWindow()

    inputfile = Path(tmp_path).joinpath("syntax_error.py")
    modify_file("print(", inputfile)
    editor = win.components["editor"]
    editor.autoreload(True)
    editor.preferences["Autoreload: watch imported modules"] = True
    editor.load_from_file(inputfile)

    win.show()
    assert win.isVisible()


code_import_module_makebox = """
from module_makebox import *
z = 1
r = makebox(z)
"""

code_module_makebox = """
import cadquery as cq
def makebox(z):
    zval = z + 1
    return cq.Workplane().box(1, 1, zval)
"""


def test_reload_import_handle_error(tmp_path, main):

    TIMEOUT = 500
    qtbot, win = main
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    traceback_view = win.components["traceback_viewer"]

    editor.autoreload(True)
    editor.preferences["Autoreload: watch imported modules"] = True

    # save the module and top level script files
    module_file = Path(tmp_path).joinpath("module_makebox.py")
    script = Path(tmp_path).joinpath("main.py")
    modify_file(code_module_makebox, module_file)
    modify_file(code_import_module_makebox, script)

    # run, verify that no exception was generated
    editor.load_from_file(script)
    debugger._actions["Run"][0].triggered.emit()
    assert traceback_view.current_exception.text() == ""

    # save the module with an error
    with qtbot.waitSignal(editor.triggerRerender, timeout=TIMEOUT):
        lines = code_module_makebox.splitlines()
        lines.remove("    zval = z + 1")  # introduce NameError
        lines = "\n".join(lines)
        modify_file(lines, module_file)

    # verify NameError is generated
    debugger._actions["Run"][0].triggered.emit()
    assert "NameError" in traceback_view.current_exception.text()

    # revert the error, verify rerender is triggered
    with qtbot.waitSignal(editor.triggerRerender, timeout=TIMEOUT):
        modify_file(code_module_makebox, module_file)

    # verify that no exception was generated
    debugger._actions["Run"][0].triggered.emit()
    assert traceback_view.current_exception.text() == ""


def test_modulefinder(tmp_path, main):

    TIMEOUT = 500
    qtbot, win = main
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    traceback_view = win.components["traceback_viewer"]
    log = win.components["log"]

    editor.autoreload(True)
    editor.preferences["Autoreload: watch imported modules"] = True

    script = Path(tmp_path).joinpath("main.py")
    Path(tmp_path).joinpath("emptydir").mkdir()
    modify_file("#import emptydir", script)
    editor.load_from_file(script)
    with qtbot.waitSignal(editor.triggerRerender, timeout=TIMEOUT):
        modify_file("import emptydir", script)

    qtbot.wait(100)
    assert "Cannot determine imported modules" in log.toPlainText().splitlines()[-1]


def test_show_all(main):

    qtbot, win = main

    editor = win.components["editor"]
    debugger = win.components["debugger"]
    object_tree = win.components["object_tree"]

    # remove all objects
    object_tree.removeObjects()
    assert object_tree.CQ.childCount() == 0

    # add code wtih Shape, Workplane, Assy, Sketch
    editor.set_text(code_show_all)

    # Run and check if all are shown
    debugger._actions["Run"][0].triggered.emit()

    assert object_tree.CQ.childCount() == 4


code_randcolor = """import cadquery as cq
b = cq.Workplane().box(8, 3, 4)
for i in range(10):
    show_object(b.translate((0,5*i,0)), options=rand_color(alpha=0))
    show_object(b.translate((0,5*i,0)), options=rand_color(0, True))
"""


def test_randcolor(main):

    qtbot, win = main

    obj_tree_comp = win.components["object_tree"]
    editor = win.components["editor"]
    debugger = win.components["debugger"]
    console = win.components["console"]

    # check that object was removed
    obj_tree_comp._toolbar_actions[0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 0

    # check that object was rendered usin explicit show_object call
    editor.set_text(code_randcolor)
    debugger._actions["Run"][0].triggered.emit()
    assert obj_tree_comp.CQ.childCount() == 2 * 10


code_show_wo_name = """
import cadquery as cq

res = cq.Workplane().box(1,1,1)

show_object(res)
show_object(cq.Workplane().box(1,1,1))
"""


def test_show_without_name(main):

    qtbot, win = main

    editor = win.components["editor"]
    debugger = win.components["debugger"]
    object_tree = win.components["object_tree"]

    # remove all objects
    object_tree.removeObjects()
    assert object_tree.CQ.childCount() == 0

    # add code wtih Shape, Workplane, Assy, Sketch
    editor.set_text(code_show_wo_name)

    # Run and check if all are shown
    debugger._actions["Run"][0].triggered.emit()

    assert object_tree.CQ.childCount() == 2

    # Check the name of the first object
    assert object_tree.CQ.child(0).text(0) == "res"

    # Check that the name of the seconf object is an int
    int(object_tree.CQ.child(1).text(0))


def test_print_redirect(main):
    qtbot, win = main

    editor = win.components["editor"]
    debugger = win.components["debugger"]
    log = win.components["log"]

    editor.set_text(r"""print("\x1b[1mfoo\x1b[0m\nbar")""")
    debugger._actions["Run"][0].triggered.emit()

    qtbot.wait(100)
    assert "foo\nbar" in log.toPlainText()


def test_log_clear(main):
    """
    Makes sure that the log can be cleared.
    """
    qtbot, win = main

    log = win.components["log"]
    log.clear_log()
    assert log.toPlainText() == ""


def test_light_dark_mode(main):
    """
    Tests that the app does switch between light and dark mode.
    """
    from PyQt5.QtGui import QPalette

    qtbot, win = main

    # Change to dark mode
    win.preferences["Light/Dark Theme"] = "Dark"
    win.preferencesChanged(None, None)

    # Retireve the toolbar so that we can check its style
    toolbar = win.toolbar

    # Get the dark mode stylesheet for the toolbar
    dark_pal = win.toolbar.palette()
    dark_bg = dark_pal.color(QPalette.Background).rgb()

    # Change to light mode
    win.preferences["Light/Dark Theme"] = "Light"
    win.preferencesChanged(None, None)

    # Get the light mode stylesheet for the toolbar
    light_pal = win.toolbar.palette()
    light_bg = light_pal.color(QPalette.Background).rgb()

    # Check that the dark mode stylesheet is different from the light mode stylesheet
    assert dark_bg != light_bg


def test_autocomplete(main):
    qtbot, win = main

    editor = win.components["editor"]

    # Set some text that should give a couple of auto-complete options
    editor.set_text(r"""import cadquery as cq\nres = cq.W""")

    # Set the cursor position to the end of the text
    editor.set_cursor_position(len(editor.get_text_with_eol()))

    # Trigger auto-complete
    editor._trigger_autocomplete()
    qtbot.wait(100)

    # Check that the completion list has two items
    assert len(editor.completion_list) == 2

    # Select the first item in the completion list
    editor.completion_list.setCurrentRow(1)

    # Wait for the completion to be applied
    qtbot.wait(100)

    # Simulate a click on the second item in the list
    editor.completion_list.itemClicked.emit(editor.completion_list.item(1))

    # Wait for the completion to be applied
    qtbot.wait(100)

    # Check that the text has been completed
    assert (
        editor.get_text_with_eol() == r"""import cadquery as cq\nres = cq.Workplane"""
    )

    # Set some text that should give a couple of auto-complete options
    editor.set_text(r"""import cadquery as cq\nres = cq.Workplane()""")

    # Set the cursor position to the end of the text
    editor.set_cursor_position(len(editor.get_text_with_eol()) - 1)

    # Trigger auto-complete
    editor._trigger_autocomplete()
    qtbot.wait(100)

    # Check to make sure that the auto-complete trigger removed the last ")"
    assert (
        editor.get_text_with_eol() == r"""import cadquery as cq\nres = cq.Workplane("""
    )


# Skip this test on Linux due to a known issue with Qt and keystrokes
@pytest.mark.skipif(
    sys.platform.startswith("linux"), reason="Known issue with Qt keystrokes"
)
def test_autocomplete_keystrokes(main):
    """
    Tests that the user keystrokes will have the intended effect on the UI.
    """

    qtbot, win = main

    editor = win.components["editor"]

    # Make sure the editor is focused
    editor.setFocus()
    qtbot.waitExposed(editor)

    # Set some text that should give a couple of auto-complete options
    editor.set_text(r"""import cadquery as cq\nres = cq.""")

    # Set the cursor position to the end of the text
    editor.set_cursor_position(len(editor.get_text_with_eol()))

    # Inject the Alt+/ key combo
    qtbot.keyClick(editor, Qt.Key_Slash, modifier=Qt.AltModifier)
    qtbot.wait(250)

    # Check that the completion list is visible
    assert editor.completion_list.isVisible()

    # Select the first item in the completion list with the Return key
    qtbot.keyClick(editor.completion_list, Qt.Key_Return)
    qtbot.wait(250)
    # Check that the text has been completed
    assert editor.get_text_with_eol() == r"""import cadquery as cq\nres = cq.Assembly"""

    # Reset for the next test
    editor.set_text(r"""import cadquery as cq\nres = cq.Workplane().box(""")

    # Set the cursor position to the end of the text
    editor.set_cursor_position(len(editor.get_text_with_eol()))

    # Inject the Alt+/ key combo
    qtbot.keyClick(editor, Qt.Key_Slash, modifier=Qt.AltModifier)
    qtbot.wait(250)

    # Check that the completion list is visible
    assert editor.completion_list.isVisible()

    # Select the first item in the completion list with the Tab key
    qtbot.keyClick(editor.completion_list, Qt.Key_Tab)
    qtbot.wait(250)

    # Check that the text has been completed
    assert (
        editor.get_text_with_eol()
        == r"""import cadquery as cq\nres = cq.Workplane().box(length,width,height,centered=True,combine=True,clean=True)"""
    )

    # Reset for the next test
    editor.set_text(r"""import cadquery as cq\nres = cq.Workplane().box(""")

    # Trigger autocomplete again
    qtbot.keyClick(editor, Qt.Key_Slash, modifier=Qt.AltModifier)
    qtbot.wait(250)

    # Check that the completion list is visible
    assert editor.completion_list.isVisible()

    # Make sure the Escape key closes the completion list
    qtbot.keyClick(editor.completion_list, Qt.Key_Escape)
    qtbot.wait(250)
    assert not editor.completion_list.isVisible()

    # Trigger autocomplete again
    qtbot.keyClick(editor, Qt.Key_Slash, modifier=Qt.AltModifier)
    qtbot.wait(250)

    # Trigger a key press that is not handled by the completion list
    qtbot.keyClick(editor.completion_list, Qt.Key_A)
    qtbot.wait(250)
    # Check that the completion list is still visible
    assert editor.completion_list.isVisible()


def test_viewer_orbit_methods(main):
    """
    Tests that mouse movements in the viewer work as expected.
    """

    qtbot, win = main

    viewer = win.components["viewer"]

    # Make sure the editor is focused
    viewer.setFocus()
    qtbot.waitExposed(viewer)

    # Simulate a drag to rotate
    qtbot.mousePress(viewer, Qt.LeftButton)
    qtbot.mouseMove(viewer, QPoint(100, 100))
    qtbot.mouseMove(viewer, QPoint(300, 300))
    qtbot.mouseRelease(viewer, Qt.LeftButton)

    # Simulate a drag to pan
    qtbot.mousePress(viewer, Qt.MiddleButton)
    event = QMouseEvent(
        QEvent.MouseMove,
        QPoint(100, 100),
        Qt.RightButton,
        Qt.RightButton,
        Qt.NoModifier,
    )
    viewer.mouseMoveEvent(event)
    event = QMouseEvent(
        QEvent.MouseMove,
        QPoint(300, 300),
        Qt.RightButton,
        Qt.RightButton,
        Qt.NoModifier,
    )
    viewer.mouseMoveEvent(event)
    qtbot.mouseRelease(viewer, Qt.MiddleButton)

    # Simulate drag to zoom
    qtbot.mousePress(viewer, Qt.RightButton)
    event = QMouseEvent(
        QEvent.MouseMove,
        QPoint(100, 100),
        Qt.RightButton,
        Qt.RightButton,
        Qt.NoModifier,
    )
    viewer.mouseMoveEvent(event)
    event = QMouseEvent(
        QEvent.MouseMove,
        QPoint(300, 300),
        Qt.RightButton,
        Qt.RightButton,
        Qt.NoModifier,
    )
    viewer.mouseMoveEvent(event)
    qtbot.mouseRelease(viewer, Qt.RightButton)

    assert True


# @pytest.mark.repeat(1)
def test_editor_autoreload(editor):

    qtbot, editor = editor

    TIMEOUT = 500

    # start out with autoreload enabled
    editor.autoreload(True)

    with open("test.py", "w") as f:
        f.write(code)

    assert editor.get_text_with_eol() == ""

    editor.load_from_file("test.py")
    assert len(editor.get_text_with_eol()) > 0

    # wait for reload.
    with qtbot.waitSignal(editor.triggerRerender, timeout=TIMEOUT):
        # modify file - NB: separate process is needed to avoid Widows quirks
        modify_file(code_bigger_object)

    # check that editor has updated file contents
    assert code_bigger_object.splitlines()[2] in editor.get_text_with_eol()

    # disable autoreload
    editor.autoreload(False)

    # Wait for reload in case it incorrectly happens. A timeout should occur
    # instead because a re-render should not be triggered with autoreload
    # disabled.
    with pytest.raises(pytestqt.exceptions.TimeoutError):
        with qtbot.waitSignal(editor.triggerRerender, timeout=TIMEOUT):
            # re-write original file contents
            modify_file(code)

    # editor should continue showing old contents since autoreload is disabled.
    assert code_bigger_object.splitlines()[2] in editor.get_text_with_eol()

    # Saving a file with autoreload disabled should not trigger a rerender.
    with pytest.raises(pytestqt.exceptions.TimeoutError):
        with qtbot.waitSignal(editor.triggerRerender, timeout=TIMEOUT):
            editor.save()

    editor.autoreload(True)

    # Saving a file with autoreload enabled should trigger a rerender.
    with qtbot.waitSignal(editor.triggerRerender, timeout=TIMEOUT):
        editor.save()


# def test_autoreload_nested(editor):

#     qtbot, editor = editor

#     TIMEOUT = 500

#     editor.autoreload(True)
#     editor.preferences["Autoreload: watch imported modules"] = True

#     with open("test_nested_top.py", "w") as f:
#         f.write(code_nested_top)

#     with open("test_nested_bottom.py", "w") as f:
#         f.write("")

#     assert editor.get_text_with_eol() == ""

#     editor.load_from_file("test_nested_top.py")
#     assert len(editor.get_text_with_eol()) > 0

#     # wait for reload.
#     with qtbot.waitSignal(editor.triggerRerender, timeout=TIMEOUT):
#         # modify file - NB: separate process is needed to avoid Windows quirks
#         modify_file(code_nested_bottom, "test_nested_bottom.py")



from PyQt5.QtWidgets import QApplication, QAction
from PyQt5.QtCore import pyqtSlot

from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager

from ..mixins import ComponentMixin

from ..icons import icon


class ConsoleWidget(RichJupyterWidget, ComponentMixin):

    name = "Console"

    def __init__(self, customBanner=None, namespace=dict(), *args, **kwargs):
        super(ConsoleWidget, self).__init__(*args, **kwargs)

        #        if not customBanner is None:
        #            self.banner = customBanner

        self._actions = {
            "Run": [
                QAction(
                    icon("clear-2"), "Clear Console", self, triggered=self.reset_console
                ),
            ]
        }
        self.font_size = 6
        self.style_sheet = """<style>
                            QPlainTextEdit, QTextEdit {
                                background-color: #3f3f3f;
                                background-clip: padding;
                                color: #dcdccc;
                                selection-background-color: #484848;
                            }
                            .inverted {
                                background-color: #dcdccc;
                                color: #3f3f3f;
                            }
                            .error { color: red; }
                            .in-prompt-number { font-weight: bold; }
                            .out-prompt-number { font-weight: bold; }
                            .in-prompt { color: navy; }
                            .out-prompt { color: darkred; }
                            </style>
                            """
        self.syntax_style = "zenburn"

        self.kernel_manager = kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel(show_banner=False)
        kernel_manager.kernel.gui = "qt"
        kernel_manager.kernel.shell.banner1 = ""

        self.kernel_client = kernel_client = self._kernel_manager.client()
        kernel_client.start_channels()

        def stop():
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()
            QApplication.instance().exit()

        self.exit_requested.connect(stop)

        self.clear()

        self.push_vars(namespace)

    @pyqtSlot(dict)
    def push_vars(self, variableDict):
        """
        Given a dictionary containing name / value pairs, push those variables
        to the Jupyter console widget
        """
        self.kernel_manager.kernel.shell.push(variableDict)

    def clear(self):
        """
        Clears the terminal
        """
        self._control.clear()

    def reset_console(self):
        """
        Resets the terminal, which clears it back to a single prompt.
        """
        self.reset(clear=True)

    def print_text(self, text):
        """
        Prints some plain text to the console
        """
        self._append_plain_text(text)

    def execute_command(self, command):
        """
        Execute a command in the frame of the console widget
        """
        self._execute(command, False)

    def _banner_default(self):

        return ""

    def app_theme_changed(self, theme):
        """
        Allows this console to be changed to match the light or dark theme of the rest of the app.
        """

        if theme == "Dark":
            self.set_default_style("linux")
        else:
            self.set_default_style("lightbg")


if __name__ == "__main__":

    import sys

    app = QApplication(sys.argv)

    console = ConsoleWidget(customBanner="IPython console test")
    console.show()

    sys.exit(app.exec_())



from PyQt5.QtWidgets import QWidget, QDialog, QTreeWidgetItem, QApplication, QAction

from PyQt5.QtCore import pyqtSlot, pyqtSignal
from PyQt5.QtGui import QIcon

from OCP.Graphic3d import (
    Graphic3d_Camera,
    Graphic3d_StereoMode,
    Graphic3d_NOM_JADE,
    Graphic3d_MaterialAspect,
)
from OCP.AIS import AIS_Shaded, AIS_WireFrame, AIS_ColoredShape, AIS_Axis
from OCP.Aspect import Aspect_GDM_Lines, Aspect_GT_Rectangular
from OCP.Quantity import (
    Quantity_NOC_BLACK as BLACK,
    Quantity_TOC_RGB as TOC_RGB,
    Quantity_Color,
)
from OCP.Geom import Geom_Axis1Placement
from OCP.gp import gp_Ax3, gp_Dir, gp_Pnt, gp_Ax1

from ..utils import layout, get_save_filename
from ..mixins import ComponentMixin
from ..icons import icon
from ..cq_utils import to_occ_color, make_AIS, DEFAULT_FACE_COLOR

from .occt_widget import OCCTWidget

from pyqtgraph.parametertree import Parameter
import qtawesome as qta


DEFAULT_EDGE_COLOR = Quantity_Color(BLACK)
DEFAULT_EDGE_WIDTH = 2


class OCCViewer(QWidget, ComponentMixin):

    name = "3D Viewer"

    preferences = Parameter.create(
        name="Pref",
        children=[
            {"name": "Fit automatically", "type": "bool", "value": True},
            {"name": "Use gradient", "type": "bool", "value": False},
            {"name": "Background color", "type": "color", "value": (95, 95, 95)},
            {"name": "Background color (aux)", "type": "color", "value": (30, 30, 30)},
            {
                "name": "Deviation",
                "type": "float",
                "value": 1e-5,
                "dec": True,
                "step": 1,
            },
            {
                "name": "Angular deviation",
                "type": "float",
                "value": 0.1,
                "dec": True,
                "step": 1,
            },
            {
                "name": "Projection Type",
                "type": "list",
                "value": "Orthographic",
                "values": [
                    "Orthographic",
                    "Perspective",
                    "Stereo",
                    "MonoLeftEye",
                    "MonoRightEye",
                ],
            },
            {
                "name": "Stereo Mode",
                "type": "list",
                "value": "QuadBuffer",
                "values": [
                    "QuadBuffer",
                    "Anaglyph",
                    "RowInterlaced",
                    "ColumnInterlaced",
                    "ChessBoard",
                    "SideBySide",
                    "OverUnder",
                ],
            },
            {
                "name": "Orbit Method",
                "type": "list",
                "value": "Turntable",
                "values": [
                    "Turntable",
                    "Trackball",
                ],
            },
        ],
    )
    IMAGE_EXTENSIONS = "png"

    sigObjectSelected = pyqtSignal(list)

    def __init__(self, parent=None):

        super(OCCViewer, self).__init__(parent)
        ComponentMixin.__init__(self)

        self.canvas = OCCTWidget()
        self.canvas.sigObjectSelected.connect(self.handle_selection)

        self.create_actions(self)

        self.layout_ = layout(
            self,
            [
                self.canvas,
            ],
            top_widget=self,
            margin=0,
        )

        self.setup_default_drawer()
        self.updatePreferences()

    def setup_default_drawer(self):

        # set the default color and material
        material = Graphic3d_MaterialAspect(Graphic3d_NOM_JADE)

        shading_aspect = self.canvas.context.DefaultDrawer().ShadingAspect()
        shading_aspect.SetMaterial(material)
        shading_aspect.SetColor(DEFAULT_FACE_COLOR)

        # face edge lw
        line_aspect = self.canvas.context.DefaultDrawer().FaceBoundaryAspect()
        line_aspect.SetWidth(DEFAULT_EDGE_WIDTH)
        line_aspect.SetColor(DEFAULT_EDGE_COLOR)

    def updatePreferences(self, *args):

        color1 = to_occ_color(self.preferences["Background color"])
        color2 = to_occ_color(self.preferences["Background color (aux)"])

        if not self.preferences["Use gradient"]:
            color2 = color1
        self.canvas.view.SetBgGradientColors(color1, color2, theToUpdate=True)

        # Set the orbit method
        orbit_method = self.preferences["Orbit Method"]
        if not orbit_method:
            orbit_method = "Trackball"
        self.canvas.set_orbit_method(orbit_method)

        self.canvas.update()

        ctx = self.canvas.context
        ctx.SetDeviationCoefficient(self.preferences["Deviation"])
        ctx.SetDeviationAngle(self.preferences["Angular deviation"])

        v = self._get_view()
        camera = v.Camera()
        projection_type = self.preferences["Projection Type"]
        camera.SetProjectionType(
            getattr(
                Graphic3d_Camera,
                f"Projection_{projection_type}",
                Graphic3d_Camera.Projection_Orthographic,
            )
        )

        # onle relevant for stereo projection
        stereo_mode = self.preferences["Stereo Mode"]
        params = v.ChangeRenderingParams()
        params.StereoMode = getattr(
            Graphic3d_StereoMode,
            f"Graphic3d_StereoMode_{stereo_mode}",
            Graphic3d_StereoMode.Graphic3d_StereoMode_QuadBuffer,
        )

    def create_actions(self, parent):

        self._actions = {
            "View": [
                QAction(
                    qta.icon("fa6s.maximize"),
                    "Fit (Shift+F1)",
                    parent,
                    shortcut="shift+F1",
                    triggered=self.fit,
                ),
                QAction(
                    QIcon(":/images/icons/isometric_view.svg"),
                    "Iso (Shift+F2)",
                    parent,
                    shortcut="shift+F2",
                    triggered=self.iso_view,
                ),
                QAction(
                    QIcon(":/images/icons/top_view.svg"),
                    "Top (Shift+F3)",
                    parent,
                    shortcut="shift+F3",
                    triggered=self.top_view,
                ),
                QAction(
                    QIcon(":/images/icons/bottom_view.svg"),
                    "Bottom (Shift+F4)",
                    parent,
                    shortcut="shift+F4",
                    triggered=self.bottom_view,
                ),
                QAction(
                    QIcon(":/images/icons/front_view.svg"),
                    "Front (Shift+F5)",
                    parent,
                    shortcut="shift+F5",
                    triggered=self.front_view,
                ),
                QAction(
                    QIcon(":/images/icons/back_view.svg"),
                    "Back (Shift+F6)",
                    parent,
                    shortcut="shift+F6",
                    triggered=self.back_view,
                ),
                QAction(
                    QIcon(":/images/icons/left_side_view.svg"),
                    "Left (Shift+F7)",
                    parent,
                    shortcut="shift+F7",
                    triggered=self.left_view,
                ),
                QAction(
                    QIcon(":/images/icons/right_side_view.svg"),
                    "Right (Shift+F8)",
                    parent,
                    shortcut="shift+F8",
                    triggered=self.right_view,
                ),
                QAction(
                    qta.icon("fa5.stop-circle"),
                    "Wireframe (Shift+F9)",
                    parent,
                    shortcut="shift+F9",
                    triggered=self.wireframe_view,
                ),
                QAction(
                    qta.icon("fa5.square"),
                    "Shaded (Shift+F10)",
                    parent,
                    shortcut="shift+F10",
                    triggered=self.shaded_view,
                ),
            ],
            "Tools": [
                QAction(
                    qta.icon("fa5s.camera"),
                    "Screenshot",
                    parent,
                    triggered=self.save_screenshot,
                )
            ],
        }

    def toolbarActions(self):

        return self._actions["View"]

    def clear(self):

        self.displayed_shapes = []
        self.displayed_ais = []
        self.canvas.context.EraseAll(True)
        context = self._get_context()
        context.PurgeDisplay()
        context.RemoveAll(True)

    def _display(self, shape):

        ais = make_AIS(shape)
        self.canvas.context.Display(shape, True)

        self.displayed_shapes.append(shape)
        self.displayed_ais.append(ais)

        # self.canvas._display.Repaint()

    @pyqtSlot(object)
    def display(self, ais):

        context = self._get_context()
        context.Display(ais, True)

        if self.preferences["Fit automatically"]:
            self.fit()

    @pyqtSlot(list)
    @pyqtSlot(list, bool)
    def display_many(self, ais_list, fit=None):
        context = self._get_context()
        for ais in ais_list:
            context.Display(ais, True)

        if self.preferences["Fit automatically"] and fit is None:
            self.fit()
        elif fit:
            self.fit()

    @pyqtSlot(QTreeWidgetItem, int)
    def update_item(self, item, col):

        ctx = self._get_context()
        if item.checkState(0):
            ctx.Display(item.ais, True)
        else:
            ctx.Erase(item.ais, True)

    @pyqtSlot(list)
    def remove_items(self, ais_items):

        ctx = self._get_context()
        for ais in ais_items:
            ctx.Erase(ais, True)

    @pyqtSlot()
    def redraw(self):

        self._get_viewer().Redraw()

    def fit(self):

        self.canvas.view.FitAll()

    def iso_view(self):

        v = self._get_view()
        v.SetProj(1, -1, 1)
        v.SetTwist(0)

    def bottom_view(self):

        v = self._get_view()
        v.SetProj(0, 0, -1)
        v.SetTwist(0)

    def top_view(self):

        v = self._get_view()
        v.SetProj(0, 0, 1)
        v.SetTwist(0)

    def front_view(self):

        v = self._get_view()
        v.SetProj(0, 1, 0)
        v.SetTwist(0)

    def back_view(self):

        v = self._get_view()
        v.SetProj(0, -1, 0)
        v.SetTwist(0)

    def left_view(self):

        v = self._get_view()
        v.SetProj(-1, 0, 0)
        v.SetTwist(0)

    def right_view(self):

        v = self._get_view()
        v.SetProj(1, 0, 0)
        v.SetTwist(0)

    def shaded_view(self):

        c = self._get_context()
        c.SetDisplayMode(AIS_Shaded, True)

    def wireframe_view(self):

        c = self._get_context()
        c.SetDisplayMode(AIS_WireFrame, True)

    def show_grid(
        self, step=1.0, size=10.0 + 1e-6, color1=(0.7, 0.7, 0.7), color2=(0, 0, 0)
    ):

        viewer = self._get_viewer()
        viewer.ActivateGrid(Aspect_GT_Rectangular, Aspect_GDM_Lines)
        viewer.SetRectangularGridGraphicValues(size, size, 0)
        viewer.SetRectangularGridValues(0, 0, step, step, 0)
        grid = viewer.Grid()
        grid.SetColors(
            Quantity_Color(*color1, TOC_RGB), Quantity_Color(*color2, TOC_RGB)
        )

    def hide_grid(self):

        viewer = self._get_viewer()
        viewer.DeactivateGrid()

    @pyqtSlot(bool, float)
    @pyqtSlot(bool)
    def toggle_grid(self, value: bool, dim: float = 10.0):

        if value:
            self.show_grid(step=dim / 20, size=dim + 1e-9)
        else:
            self.hide_grid()

    @pyqtSlot(gp_Ax3)
    def set_grid_orientation(self, orientation: gp_Ax3):

        viewer = self._get_viewer()
        viewer.SetPrivilegedPlane(orientation)

    def show_axis(self, origin=(0, 0, 0), direction=(0, 0, 1)):

        ax_placement = Geom_Axis1Placement(gp_Ax1(gp_Pnt(*origin), gp_Dir(*direction)))
        ax = AIS_Axis(ax_placement)
        self._display_ais(ax)

    def save_screenshot(self):

        fname = get_save_filename(self.IMAGE_EXTENSIONS)
        if fname != "":
            self._get_view().Dump(fname)

    def _display_ais(self, ais):

        self._get_context().Display(ais)

    def _get_view(self):

        return self.canvas.view

    def _get_viewer(self):

        return self.canvas.viewer

    def _get_context(self):

        return self.canvas.context

    @pyqtSlot(list)
    def handle_selection(self, obj):

        self.sigObjectSelected.emit(obj)

    @pyqtSlot(list)
    def set_selected(self, ais):

        ctx = self._get_context()
        ctx.ClearSelected(False)

        for obj in ais:
            ctx.AddOrRemoveSelected(obj, False)

        self.redraw()


if __name__ == "__main__":

    import sys
    from OCP.BRepPrimAPI import BRepPrimAPI_MakeBox

    app = QApplication(sys.argv)
    viewer = OCCViewer()

    dlg = QDialog()
    dlg.setFixedHeight(400)
    dlg.setFixedWidth(600)

    layout(dlg, (viewer,), dlg)
    dlg.show()

    box = BRepPrimAPI_MakeBox(20, 20, 30)
    box_ais = AIS_ColoredShape(box.Shape())
    viewer.display(box_ais)

    sys.exit(app.exec_())



from PyQt5.QtWidgets import (
    QTreeWidget,
    QTreeWidgetItem,
    QAction,
    QMenu,
    QWidget,
    QAbstractItemView,
)
from PyQt5.QtCore import Qt, pyqtSlot, pyqtSignal

from pyqtgraph.parametertree import Parameter, ParameterTree

from OCP.AIS import AIS_Line
from OCP.Geom import Geom_Line
from OCP.gp import gp_Dir, gp_Pnt, gp_Ax1

from ..mixins import ComponentMixin
from ..icons import icon
from ..cq_utils import (
    make_AIS,
    export,
    to_occ_color,
    is_obj_empty,
    get_occ_color,
    set_color,
)
from .viewer import DEFAULT_FACE_COLOR
from ..utils import splitter, layout, get_save_filename


class TopTreeItem(QTreeWidgetItem):

    def __init__(self, *args, **kwargs):

        super(TopTreeItem, self).__init__(*args, **kwargs)


class ObjectTreeItem(QTreeWidgetItem):

    props = [
        {"name": "Name", "type": "str", "value": "", "readonly": True},
        # {"name": "Color", "type": "color", "value": "#f4a824"},
        # {"name": "Alpha", "type": "float", "value": 0, "limits": (0, 1), "step": 1e-1},
        {"name": "Visible", "type": "bool", "value": True},
    ]

    def __init__(
        self,
        name,
        ais=None,
        shape=None,
        shape_display=None,
        sig=None,
        alpha=0.0,
        color="#f4a824",
        **kwargs,
    ):

        super(ObjectTreeItem, self).__init__([name], **kwargs)
        self.setFlags(self.flags() | Qt.ItemIsUserCheckable)
        self.setCheckState(0, Qt.Checked)

        self.ais = ais
        self.shape = shape
        self.shape_display = shape_display
        self.sig = sig

        self.properties = Parameter.create(name="Properties", children=self.props)

        self.properties["Name"] = name
        # Alpha and Color from this panel fight with the options in show_object and so they are
        # disabled for now until a better solution is found
        # self.properties["Alpha"] = ais.Transparency()
        # self.properties["Color"] = (
        #     get_occ_color(ais)
        #     if ais and ais.HasColor()
        #     else get_occ_color(DEFAULT_FACE_COLOR)
        # )
        self.properties.sigTreeStateChanged.connect(self.propertiesChanged)

    def propertiesChanged(self, properties, changed):

        changed_prop = changed[0][0]

        self.setData(0, 0, self.properties["Name"])

        # if changed_prop.name() == "Alpha":
        #     self.ais.SetTransparency(self.properties["Alpha"])

        # if changed_prop.name() == "Color":
        #     set_color(self.ais, to_occ_color(self.properties["Color"]))

        # self.ais.Redisplay()

        if self.properties["Visible"]:
            self.setCheckState(0, Qt.Checked)
        else:
            self.setCheckState(0, Qt.Unchecked)

        if self.sig:
            self.sig.emit()


class CQRootItem(TopTreeItem):

    def __init__(self, *args, **kwargs):

        super(CQRootItem, self).__init__(["CQ models"], *args, **kwargs)


class HelpersRootItem(TopTreeItem):

    def __init__(self, *args, **kwargs):

        super(HelpersRootItem, self).__init__(["Helpers"], *args, **kwargs)


class ObjectTree(QWidget, ComponentMixin):

    name = "Object Tree"
    _stash = []

    preferences = Parameter.create(
        name="Preferences",
        children=[
            {"name": "Preserve properties on reload", "type": "bool", "value": False},
            {"name": "Clear all before each run", "type": "bool", "value": True},
            {"name": "STL precision", "type": "float", "value": 0.1},
        ],
    )

    sigObjectsAdded = pyqtSignal([list], [list, bool])
    sigObjectsRemoved = pyqtSignal(list)
    sigCQObjectSelected = pyqtSignal(object)
    sigAISObjectsSelected = pyqtSignal(list)
    sigItemChanged = pyqtSignal(QTreeWidgetItem, int)
    sigObjectPropertiesChanged = pyqtSignal()

    def __init__(self, parent):

        super(ObjectTree, self).__init__(parent)

        self.tree = tree = QTreeWidget(
            self, selectionMode=QAbstractItemView.ExtendedSelection
        )
        self.properties_editor = ParameterTree(self)

        tree.setHeaderHidden(True)
        tree.setItemsExpandable(False)
        tree.setRootIsDecorated(False)
        tree.setContextMenuPolicy(Qt.ActionsContextMenu)

        # forward itemChanged singal
        tree.itemChanged.connect(lambda item, col: self.sigItemChanged.emit(item, col))
        # handle visibility changes form tree
        tree.itemChanged.connect(self.handleChecked)

        self.CQ = CQRootItem()
        self.Helpers = HelpersRootItem()

        root = tree.invisibleRootItem()
        root.addChild(self.CQ)
        root.addChild(self.Helpers)

        tree.expandToDepth(1)

        self._export_STL_action = QAction(
            "Export as STL",
            self,
            enabled=False,
            triggered=lambda: self.export("stl", self.preferences["STL precision"]),
        )

        self._export_STEP_action = QAction(
            "Export as STEP", self, enabled=False, triggered=lambda: self.export("step")
        )

        self._clear_current_action = QAction(
            icon("delete"),
            "Clear current",
            self,
            enabled=False,
            triggered=self.removeSelected,
        )

        self._toolbar_actions = [
            QAction(
                icon("delete-many"), "Clear all", self, triggered=self.removeObjects
            ),
            self._clear_current_action,
        ]

        self.prepareMenu()

        tree.itemSelectionChanged.connect(self.handleSelection)
        tree.customContextMenuRequested.connect(self.showMenu)

        self.prepareLayout()

    def prepareMenu(self):

        self.tree.setContextMenuPolicy(Qt.CustomContextMenu)

        self._context_menu = QMenu(self)
        self._context_menu.addActions(self._toolbar_actions)
        self._context_menu.addActions(
            (self._export_STL_action, self._export_STEP_action)
        )

    def prepareLayout(self):

        self._splitter = splitter(
            (self.tree, self.properties_editor),
            stretch_factors=(2, 1),
            orientation=Qt.Vertical,
        )
        layout(self, (self._splitter,), top_widget=self)

        self._splitter.show()

    def showMenu(self, position):

        self._context_menu.exec_(self.tree.viewport().mapToGlobal(position))

    def menuActions(self):

        return {"Tools": [self._export_STL_action, self._export_STEP_action]}

    def toolbarActions(self):

        return self._toolbar_actions

    def addLines(self):

        origin = (0, 0, 0)
        ais_list = []

        for name, color, direction in zip(
            ("X", "Y", "Z"),
            ("red", "lawngreen", "blue"),
            ((1, 0, 0), (0, 1, 0), (0, 0, 1)),
        ):
            line_placement = Geom_Line(gp_Ax1(gp_Pnt(*origin), gp_Dir(*direction)))
            line = AIS_Line(line_placement)
            line.SetColor(to_occ_color(color))

            self.Helpers.addChild(ObjectTreeItem(name, ais=line))

            ais_list.append(line)

        self.sigObjectsAdded.emit(ais_list)

    def _current_properties(self):

        current_params = {}
        for i in range(self.CQ.childCount()):
            child = self.CQ.child(i)
            current_params[child.properties["Name"]] = child.properties

        return current_params

    def _restore_properties(self, obj, properties):

        for p in properties[obj.properties["Name"]]:
            obj.properties[p.name()] = p.value()

    @pyqtSlot(dict, bool)
    @pyqtSlot(dict)
    def addObjects(self, objects, clean=False, root=None):

        if root is None:
            root = self.CQ

        request_fit_view = True if root.childCount() == 0 else False
        preserve_props = self.preferences["Preserve properties on reload"]

        if preserve_props:
            current_props = self._current_properties()

        if clean or self.preferences["Clear all before each run"]:
            self.removeObjects()

        ais_list = []

        # remove empty objects
        objects_f = {k: v for k, v in objects.items() if not is_obj_empty(v.shape)}

        for name, obj in objects_f.items():
            ais, shape_display = make_AIS(obj.shape, obj.options)

            child = ObjectTreeItem(
                name,
                shape=obj.shape,
                shape_display=shape_display,
                ais=ais,
                sig=self.sigObjectPropertiesChanged,
            )

            if preserve_props and name in current_props:
                self._restore_properties(child, current_props)

            if child.properties["Visible"]:
                ais_list.append(ais)

            root.addChild(child)

        if request_fit_view:
            self.sigObjectsAdded[list, bool].emit(ais_list, True)
        else:
            self.sigObjectsAdded[list].emit(ais_list)

    @pyqtSlot(object, str, object)
    def addObject(self, obj, name="", options=None):

        if options is None:
            options = {}

        root = self.CQ

        ais, shape_display = make_AIS(obj, options)

        root.addChild(
            ObjectTreeItem(
                name,
                shape=obj,
                shape_display=shape_display,
                ais=ais,
                sig=self.sigObjectPropertiesChanged,
            )
        )

        self.sigObjectsAdded.emit([ais])

    @pyqtSlot(list)
    @pyqtSlot()
    def removeObjects(self, objects=None):

        if objects:
            removed_items_ais = [self.CQ.takeChild(i).ais for i in objects]
        else:
            removed_items_ais = [ch.ais for ch in self.CQ.takeChildren()]

        self.sigObjectsRemoved.emit(removed_items_ais)

    @pyqtSlot(bool)
    def stashObjects(self, action: bool):

        if action:
            self._stash = self.CQ.takeChildren()
            removed_items_ais = [ch.ais for ch in self._stash]
            self.sigObjectsRemoved.emit(removed_items_ais)
        else:
            self.removeObjects()
            self.CQ.addChildren(self._stash)
            ais_list = [el.ais for el in self._stash]
            self.sigObjectsAdded.emit(ais_list)

    @pyqtSlot()
    def removeSelected(self):

        ixs = self.tree.selectedIndexes()
        rows = [ix.row() for ix in ixs]

        self.removeObjects(rows)

    def export(self, export_type, precision=None):

        items = self.tree.selectedItems()

        # if CQ models is selected get all children
        if [item for item in items if item is self.CQ]:
            CQ = self.CQ
            shapes = [CQ.child(i).shape for i in range(CQ.childCount())]
        # otherwise collect all selected children of CQ
        else:
            shapes = [item.shape for item in items if item.parent() is self.CQ]

        fname = get_save_filename(export_type)
        if fname != "":
            export(shapes, export_type, fname, precision)

    @pyqtSlot()
    def handleSelection(self):

        items = self.tree.selectedItems()
        if len(items) == 0:
            self._export_STL_action.setEnabled(False)
            self._export_STEP_action.setEnabled(False)
            return

        # emit list of all selected ais objects (might be empty)
        ais_objects = [item.ais for item in items if item.parent() is self.CQ]
        self.sigAISObjectsSelected.emit(ais_objects)

        # handle context menu and emit last selected CQ  object (if present)
        item = items[-1]
        if item.parent() is self.CQ:
            self._export_STL_action.setEnabled(True)
            self._export_STEP_action.setEnabled(True)
            self._clear_current_action.setEnabled(True)
            self.sigCQObjectSelected.emit(item.shape)
            self.properties_editor.setParameters(item.properties, showTop=False)
            self.properties_editor.setEnabled(True)
        elif item is self.CQ and item.childCount() > 0:
            self._export_STL_action.setEnabled(True)
            self._export_STEP_action.setEnabled(True)
        else:
            self._export_STL_action.setEnabled(False)
            self._export_STEP_action.setEnabled(False)
            self._clear_current_action.setEnabled(False)
            self.properties_editor.setEnabled(False)
            self.properties_editor.clear()

    @pyqtSlot(list)
    def handleGraphicalSelection(self, shapes):

        self.tree.clearSelection()

        CQ = self.CQ
        for i in range(CQ.childCount()):
            item = CQ.child(i)
            for shape in shapes:
                if item.ais.Shape().IsEqual(shape):
                    item.setSelected(True)

    @pyqtSlot(QTreeWidgetItem, int)
    def handleChecked(self, item, col):

        if type(item) is ObjectTreeItem:
            if item.checkState(0):
                item.properties["Visible"] = True
            else:
                item.properties["Visible"] = False



import requests

from packaging.version import parse as parse_version

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtCore import QUrl
from PyQt5.QtWidgets import QFileDialog, QMessageBox

DOCK_POSITIONS = {
    "right": QtCore.Qt.RightDockWidgetArea,
    "left": QtCore.Qt.LeftDockWidgetArea,
    "top": QtCore.Qt.TopDockWidgetArea,
    "bottom": QtCore.Qt.BottomDockWidgetArea,
}


def layout(
    parent,
    items,
    top_widget=None,
    layout_type=QtWidgets.QVBoxLayout,
    margin=2,
    spacing=0,
):

    if not top_widget:
        top_widget = QtWidgets.QWidget(parent)
        top_widget_was_none = True
    else:
        top_widget_was_none = False
    layout = layout_type(top_widget)
    top_widget.setLayout(layout)

    for item in items:
        layout.addWidget(item)

    layout.setSpacing(spacing)
    layout.setContentsMargins(margin, margin, margin, margin)

    if top_widget_was_none:
        return top_widget
    else:
        return layout


def splitter(items, stretch_factors=None, orientation=QtCore.Qt.Horizontal):

    sp = QtWidgets.QSplitter(orientation)

    for item in items:
        sp.addWidget(item)

    if stretch_factors:
        for i, s in enumerate(stretch_factors):
            sp.setStretchFactor(i, s)

    return sp


def dock(
    widget,
    title,
    parent,
    allowedAreas=QtCore.Qt.AllDockWidgetAreas,
    defaultArea="right",
    name=None,
    icon=None,
):

    dock = QtWidgets.QDockWidget(title, parent, objectName=title)

    if name:
        dock.setObjectName(name)
    if icon:
        dock.toggleViewAction().setIcon(icon)

    dock.setAllowedAreas(allowedAreas)
    dock.setWidget(widget)
    action = dock.toggleViewAction()
    action.setText(title)

    dock.setFeatures(
        QtWidgets.QDockWidget.DockWidgetFeatures(
            QtWidgets.QDockWidget.AllDockWidgetFeatures
        )
    )

    parent.addDockWidget(DOCK_POSITIONS[defaultArea], dock)

    return dock


def add_actions(menu, actions):

    if len(actions) > 0:
        menu.addActions(actions)
        menu.addSeparator()


def open_url(url):

    QDesktopServices.openUrl(QUrl(url))


def about_dialog(parent, title, text):

    QtWidgets.QMessageBox.about(parent, title, text)


def get_save_filename(suffix):

    rv, _ = QFileDialog.getSaveFileName(filter="*.{}".format(suffix))
    if rv != "" and not rv.endswith(suffix):
        rv += "." + suffix

    return rv


def get_open_filename(suffix, curr_dir):

    rv, _ = QFileDialog.getOpenFileName(
        directory=curr_dir, filter="*.{}".format(suffix)
    )
    if rv != "" and not rv.endswith(suffix):
        rv += "." + suffix

    return rv


def check_gtihub_for_updates(
    parent, mod, github_org="cadquery", github_proj="cadquery"
):

    url = f"https://api.github.com/repos/{github_org}/{github_proj}/releases"
    resp = requests.get(url).json()

    newer = [
        el["tag_name"]
        for el in resp
        if not el["draft"]
        and parse_version(el["tag_name"]) > parse_version(mod.__version__)
    ]

    if newer:
        title = "Updates available"
        text = (
            f"There are newer versions of {github_proj} "
            f"available on github:\n" + "\n".join(newer)
        )

    else:
        title = "No updates available"
        text = f"You are already using the latest version of {github_proj}"

    QtWidgets.QMessageBox.about(parent, title, text)


def confirm(parent, title, msg):

    rv = QMessageBox.question(parent, title, msg, QMessageBox.Yes, QMessageBox.No)

    return True if rv == QMessageBox.Yes else False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 23 22:02:30 2018

@author: adam
"""

from functools import reduce
from operator import add
from logbook import Logger

from PyQt5.QtCore import pyqtSlot, QSettings


class MainMixin(object):

    name = "Main"
    org = "Unknown"

    components = {}
    docks = {}
    preferences = None

    def __init__(self):

        self.settings = QSettings(self.org, self.name)

    def registerComponent(self, name, component, dock=None):

        self.components[name] = component

        if dock:
            self.docks[name] = dock(component)

    def saveWindow(self):

        self.settings.setValue("geometry", self.saveGeometry())
        self.settings.setValue("windowState", self.saveState())

    def restoreWindow(self):

        if self.settings.value("geometry"):
            self.restoreGeometry(self.settings.value("geometry"))
        if self.settings.value("windowState"):
            self.restoreState(self.settings.value("windowState"))

    def savePreferences(self):

        settings = self.settings

        if self.preferences:
            settings.setValue("General", self.preferences.saveState())

        for comp in (c for c in self.components.values() if c.preferences):
            settings.setValue(comp.name, comp.preferences.saveState())

    def restorePreferences(self):

        settings = self.settings

        if self.preferences and settings.value("General"):
            self.preferences.restoreState(
                settings.value("General"), removeChildren=False
            )

        for comp in (c for c in self.components.values() if c.preferences):
            if settings.value(comp.name):
                comp.preferences.restoreState(
                    settings.value(comp.name), removeChildren=False
                )

    def saveComponentState(self):

        settings = self.settings

        for comp in self.components.values():
            comp.saveComponentState(settings)

    def restoreComponentState(self):

        settings = self.settings

        for comp in self.components.values():
            comp.restoreComponentState(settings)


class ComponentMixin(object):

    name = "Component"
    preferences = None

    _actions = {}

    def __init__(self):

        if self.preferences:
            self.preferences.sigTreeStateChanged.connect(self.updatePreferences)

        self._logger = Logger(self.name)

    def menuActions(self):

        return self._actions

    def toolbarActions(self):

        if len(self._actions) > 0:
            return reduce(add, [a for a in self._actions.values()])
        else:
            return []

    @pyqtSlot(object, object)
    def updatePreferences(self, *args):

        pass

    def saveComponentState(self, store):

        pass

    def restoreComponentState(self, store):

        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 25 14:47:10 2018

@author: adam
"""

from PyQt5.QtGui import QIcon

from . import icons_res

_icons = {"app": QIcon(":/images/icons/cadquery_logo_dark.svg")}

import qtawesome as qta

_icons_specs = {
    "new": (("fa5.file",), {}),
    "open": (("fa5.folder-open",), {}),
    # borrowed from spider-ide
    "autoreload": [
        ("fa5s.redo-alt", "fa5.clock"),
        {
            "options": [
                {"scale_factor": 0.75, "offset": (-0.1, -0.1)},
                {"scale_factor": 0.5, "offset": (0.25, 0.25)},
            ]
        },
    ],
    "save": (("fa5.save",), {}),
    "save_as": (
        ("fa5.save", "fa5s.pencil-alt"),
        {
            "options": [
                {
                    "scale_factor": 1,
                },
                {"scale_factor": 0.8, "offset": (0.2, 0.2)},
            ]
        },
    ),
    "run": (("fa5s.play",), {}),
    "debug": (("fa5s.bug",), {}),
    "delete": (("fa5s.trash",), {}),
    "delete-many": (
        (
            "fa5s.trash",
            "fa5s.trash",
        ),
        {
            "options": [
                {"scale_factor": 0.8, "offset": (0.2, 0.2), "color": "gray"},
                {"scale_factor": 0.8},
            ]
        },
    ),
    "help": (("fa5s.life-ring",), {}),
    "about": (("fa5s.info",), {}),
    "preferences": (("fa5s.cogs",), {}),
    "inspect": (
        ("fa5s.cubes", "fa5s.search"),
        {"options": [{"scale_factor": 0.8, "offset": (0, 0), "color": "gray"}, {}]},
    ),
    "screenshot": (("fa5s.camera",), {}),
    "screenshot-save": (
        ("fa5.save", "fa5s.camera"),
        {
            "options": [
                {"scale_factor": 0.8},
                {"scale_factor": 0.8, "offset": (0.2, 0.2)},
            ]
        },
    ),
    "toggle-comment": (("fa5s.hashtag",), {}),
    "search": (("fa5s.search",), {}),
    "arrow-step-over": (("fa5s.step-forward",), {}),
    "arrow-step-in": (("fa5s.angle-down",), {}),
    "arrow-continue": (("fa5s.arrow-right",), {}),
    "clear": (("fa5s.eraser",), {}),
    "clear-2": (("fa5s.broom",), {}),
}


def icon(name):

    if name in _icons:
        return _icons[name]

    args, kwargs = _icons_specs[name]

    return qta.icon(*args, **kwargs)



from traceback import extract_tb, format_exception_only
from itertools import dropwhile

from PyQt5.QtWidgets import QWidget, QTreeWidget, QTreeWidgetItem, QAction, QLabel
from PyQt5.QtCore import Qt, pyqtSlot, pyqtSignal
from PyQt5.QtGui import QFontMetrics

from ..mixins import ComponentMixin
from ..utils import layout


class TracebackTree(QTreeWidget):

    name = "Traceback Viewer"

    def __init__(self, parent):

        super(TracebackTree, self).__init__(parent)
        self.setHeaderHidden(False)
        self.setItemsExpandable(False)
        self.setRootIsDecorated(False)
        self.setContextMenuPolicy(Qt.ActionsContextMenu)

        self.setColumnCount(3)
        self.setHeaderLabels(["File", "Line", "Code"])

        self.root = self.invisibleRootItem()


class TracebackPane(QWidget, ComponentMixin):

    sigHighlightLine = pyqtSignal(int)

    def __init__(self, parent):

        super(TracebackPane, self).__init__(parent)

        self.tree = TracebackTree(self)
        self.current_exception = QLabel(self)
        self.current_exception.setStyleSheet("QLabel {color : red; }")

        layout(self, (self.current_exception, self.tree), self)

        self.tree.currentItemChanged.connect(self.handleSelection)

    def truncate_text(self, text, max_length=100):
        """
        Used to prevent the label from expanding the window width off the screen.
        """
        metrics = QFontMetrics(self.current_exception.font())
        elided_text = metrics.elidedText(
            text, Qt.ElideRight, self.current_exception.width() - 75
        )

        return elided_text

    @pyqtSlot(object, str)
    def addTraceback(self, exc_info, code):

        self.tree.clear()

        if exc_info:
            t, exc, tb = exc_info

            root = self.tree.root
            code = code.splitlines()

            for el in dropwhile(
                lambda el: "string>" not in el.filename, extract_tb(tb)
            ):
                # workaround of the traceback module
                if el.line == "":
                    line = code[el.lineno - 1].strip()
                else:
                    line = el.line

                root.addChild(QTreeWidgetItem([el.filename, str(el.lineno), line]))

            exc_name = t.__name__
            exc_msg = str(exc)
            exc_msg = exc_msg.replace("<", "&lt;").replace(">", "&gt;")  # replace <>

            truncated_msg = self.truncate_text(exc_msg)
            self.current_exception.setText(
                "<b>{}</b>: {}".format(exc_name, truncated_msg)
            )
            self.current_exception.setToolTip(exc_msg)

            # handle the special case of a SyntaxError
            if t is SyntaxError:
                root.addChild(
                    QTreeWidgetItem(
                        [
                            exc.filename,
                            str(exc.lineno),
                            exc.text.strip() if exc.text else "",
                        ]
                    )
                )
        else:
            self.current_exception.setText("")
            self.current_exception.setToolTip("")

    @pyqtSlot(QTreeWidgetItem, QTreeWidgetItem)
    def handleSelection(self, item, *args):

        if item:
            f, line = item.data(0, 0), int(item.data(1, 0))

            if "<string>" in f:
                self.sigHighlightLine.emit(line)
